
	/* assign unique IDs to all stars, based on their aligned positions */

	/***********************************************************************
	 * Changes:
	 *
	 *  - fixed off-by-one error in "get_star_index".
	 *      MWR  12/14/2002
	 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <search.h>
#include "multipht.h"

#undef DEBUG

#define HIST_BINS    20			/* number of bins in histogram */

static struct s_hash *hashtab;		/* hash table array */
static int nhash = 0;				/* number of hash table entries */

static void sort_stars(struct s_image *im);
static int comp_by_xc(struct s_star *s1, struct s_star *s2);
struct s_hash * lookup(struct s_star *star);
static struct s_hash * looklist(struct s_star *star, int index);
static void hash_insert(struct s_star *star);
static void hash_add(struct s_star *star, struct s_hash *p);
static void add_id(struct s_hash *p);
static void init_hash(void);
static int is_a_match(struct s_star *s1, struct s_hash *p);
static int get_star_index(struct s_star *starp);
#ifdef DEBUG
static void print_hash(void);
static void print_by_id(void);
static int verify_table(void);
#endif

	/**************************************************************
	 * PROCEDURE: good_images
	 *
	 * DESCRIPTION: Assign unique IDs to each star in each image, 
	 *              based upon its position in the template coordinate system. 
	 *              Note that the stars should already be almost - but not
	 *              quite - sorted by their rows, 
	 *              so we have to sort them first.  
	 *
	 *              As each star is examined, the 'id' field of 
	 *              its structure is set the master ID number for all 
	 *              stars in all images.  Moreover, the hash table entry 
	 *              for that ID has the new star added to its linked 
	 *              list of stars that share the ID, and has its 
	 *              'num' field incremented.  thus, we can go 
	 *              through all stars BY ID efficiently after having 
	 *              built up the table, and we can quickly find out how
	 *              many stars share the same ID 
	 *              (i.e. in how many images was each star found).
	 *
	 * RETURNS:
	 *     the total number of unique stars found 
	 */

int
assign_ids
	(
	void
	)
{
	int i, j;
	struct s_hash *p; 
	struct s_star *s;

	init_hash();
	for (i = 0; i < num_images; i++) {
		sort_stars(&(image_arr[i]));
		for (j = 0; j < image_arr[i].num_star; j++) {
			s = &(image_arr[i].stars[j]);
			if ((p = lookup(s)) != NULL) {
				hash_add(s, p);
			}
			else {
				hash_insert(s);
			}
#ifdef DEBUG
			printf("adding entry for image %6d star %5d \n", i, j);
			verify_table();
#endif
		}
	}

	if (debug_flag) {
#ifdef DEBUG
		print_by_id();
		print_hash();
		print_star_hist();
		print_image_hist();
#endif
	}

	return(num_ids);
}

	/**************************************************************
	 * PROCEDURE: sort_stars
	 *
	 * DESCRIPTION: Sort all the stars in the given image by 
	 *              their 'xc' value.
	 *
	 * RETURNS:
	 *     nothing
	 */


static void
sort_stars
	(
   struct s_image *im
	)
{
	qsort((char *) &(im->stars[0]), im->num_star, sizeof(struct s_star),
				(PFI) comp_by_xc);
}


	/**************************************************************
	 * PROCEDURE: comp_by_xc
	 *
	 * DESCRIPTION: Compare the 'xc' field of two star structures.
	 *
	 * RETURNS:
	 *     -1 if the first star has an 'xc' coord less than second
	 *      0                                     equal to 
	 *     +1                                     greater than 
	 */

static int
comp_by_xc
	(
	struct s_star *s1, 
	struct s_star *s2
	)
{
	if (s1->xc < s2->xc) {
		return(-1);
	}
	else if (s1->xc > s2->xc) {
		return(1);;
	}
	else {
		return(0);
	}
}
	

	/**************************************************************
	 * PROCEDURE: lookup
	 *
	 * DESCRIPTION: Lookup a star in the hash table, 
	 *              using its 'xc' coord as a key.
	 *
	 * RETURNS:
	 *      a pointer to the hash table entry, if found
	 *      NULL,                              if not found. 
	 */


struct s_hash *
lookup
	(
	struct s_star *star
	)
{
	int i, index;
	struct s_hash *found;

	index = get_star_index(star);

	/* look first in the linked list of stars with the same 'xc' value */
	if ((found = looklist(star, index)) != (struct s_hash *) NULL) {
		return(found);
	}
	
	/* now look in the linked lists of stars with progressively smaller xc
	   values, down to 'xc - 2*RAD'. */
	for (i = index - 1; (i >= 0) && (i >= index - (int)(2*RADIUS)); i--) {
		if ((found = looklist(star, i)) != (struct s_hash *) NULL) {
			return(found);
		}
	}

	/* now look in the linked lists of stars with progressively larger xc
	   values, up to 'xc + RAD'. */
	for (i = index + 1; (i < nhash) && (i <= index + (int)(2*RADIUS)); i++) {
		if ((found = looklist(star, i)) != (struct s_hash *) NULL) {
			return(found);
		}
	}

	return((struct s_hash *) NULL);
}


	/**************************************************************
	 * PROCEDURE: looklist
	 *
	 * DESCRIPTION: Look for a star in the given linked list of 
	 *              hash entries.  
	 *
	 * RETURNS:
	 *     a pointer to the has entry, if the star is found
	 *     NULL                        if the star is not found
	 */

static struct s_hash *
looklist
	(
	struct s_star *star,
	int index
	)
{
	struct s_hash *p;

	if ((index < 0) || (index >= nhash)) {
		return((struct s_hash *) NULL);
	}
	for (p = &(hashtab[index]); (p != NULL) && (p->id != -1); p = p->next) {
		if (is_a_match(star, p)) {
			return(p);
		}
	}

	return((struct s_hash *) NULL);
}



	/************************************************************************8
	 * PROCEDURE: verify_table
	 *
	 * DESCRIPTION: walk through the hash table, and all its entries,
	 *              checking the pointer values.
	 *
	 *              Used only in debugging.
	 *
	 * RETURNS:
	 *   0          if all goes well
	 *   1          if a problem is found
	 */

#ifdef DEBUG
static int
verify_table
	(
	void
	)
{
	int i, j;
	struct s_hash *hash_p;
	struct s_starp *starp_p;

	for (i = 0; i < nhash; i++) {
		hash_p = &(hashtab[i]);
		if (hash_p != NULL) {
			/* check all the stars with this ID */
			starp_p = hash_p->starp;
			for (j = 0; j < hash_p->num; j++) {
				if (starp_p == NULL) {
					printf("verify_table: hashtab[%d] star %d has null pointer\n",
							      i, j);
					return(1);
				}
				starp_p = starp_p->next;
			}
			if ((hash_p->num > 0) && (starp_p != NULL)) {
				printf("verify_table: hashtab[%d] starlist missing NULL starp->next\n", 
								i);
			}
			/* now check all other hash entries with identical has value */
			while (hash_p->next != NULL) {
				hash_p = hash_p->next;
				/* check _this_ entry's stars */
				starp_p = hash_p->starp;
				for (j = 0; j < hash_p->num; j++) {
					if (starp_p == NULL) {
						printf("verify_table: hashtab[%d] XX star %d has null pointer\n",
								      i, j);
						return(1);
					}
					starp_p = starp_p->next;
				}
			}
		}
	}

	return(0);
}
#endif


	/**************************************************************
	 * PROCEDURE: hash_insert
	 *
	 * DESCRIPTION: Insert a star into the hash table
	 *
	 * RETURNS:
	 *     nothing
	 */

static void
hash_insert
	(
	struct s_star *star
	)
{
	int index, flag;
	struct s_hash *p, *q;

	index = get_star_index(star);
	if (hashtab[index].id == -1) {
		flag = 1;
	}
	else {
		flag = 0;
	}

	if (flag == 1) {
		p = &(hashtab[index]);
	}
	else {
		if ((p = (struct s_hash *) malloc(sizeof(struct s_hash))) == NULL) {
			fprintf(stderr, "hash_insert: can't malloc for new entry, num_ids=%d\n",
					num_ids);
			exit(-1);
		}
		p->num = 0;
	}
	p->xc = star->xc;
	p->yc = star->yc;
	p->id = num_ids++;
	star->id = p->id;
	p->num++;
	if ((p->starp = (struct s_starp *) malloc(sizeof(struct s_starp))) == NULL) {
		fprintf(stderr, "hash_insert: can't malloc for starp, num_ids=%d\n", 
							num_ids);
		exit(-1);
	}
	p->starp->star = star;
	p->starp->next = (struct s_starp *) NULL;
	p->next = (struct s_hash *) NULL;
	p->include = 1;

	if (flag == 0) {
		/* walk down the linked list of hash structures to the end, then
		   put the new structure at the end of the list */	
		for (q = &(hashtab[index]); q->next != NULL; q = q->next) {
			;
		}
		q->next = p;
	}

	add_id(p);

#ifdef DEBUG
	if (p->starp->next == NULL) {
		printf("hash_insert: after inserting index %6d, starp->next is NULL\n",
				index);
	}
	else {
		printf("hash_insert: after inserting index %6d, starp->next bad??\n",
				index);
	}
#endif

}
	

	/**************************************************************
	 * PROCEDURE: hash_add
	 *
	 * DESCRIPTION: Add a star to the hash entry that matches it.  
	 *              Modify the hash entry's 'xc' and 'yc' values so 
	 *              that they are the average of all matching star values. 
	 *
	 * RETURNS:
	 *     nothing
	 */

static void
hash_add
	(
	struct s_star *star,
	struct s_hash *p
	)
{
	double dn;
	struct s_starp *sp, *q;

#ifdef DEBUG
	printf("hash_add: about to add star for hash %d \n", p->id);
#endif

	if ((sp = (struct s_starp *) malloc(sizeof(struct s_starp))) == NULL) {
		fprintf(stderr, 
                  "hash_add: can't malloc for starp, num_ids=%d\n", num_ids);
		exit(-1);
	}
	sp->star = star;
	sp->next = (struct s_starp *) NULL;
	p->num++;
	star->id = p->id;
	for (q = p->starp; (q != NULL) && (q->next != NULL); q = q->next) {
		;
	}
	q->next = sp;
	
	/* now modify the hash entry's xc and yc values */
	dn = p->num;
	p->xc = ((p->xc*dn) + star->xc)/(dn + 1.0);
	p->yc = ((p->yc*dn) + star->yc)/(dn + 1.0);
}	



	/**************************************************************
	 * PROCEDURE: add_id
	 *
	 * DESCRIPTION: Add a new entry to the linked list of hash entries 
	 *              which are in order by their ID numbers.  
	 *              The head of this list is 'star_list', and by walking 
	 *              down it, you can walk through all hash entries 
	 *              (and, hence, stars) by their ID number 
	 *
	 * RETURNS:
	 *     nothing
	 */

static void
add_id
	(
	struct s_hash *p
	)
{
	static int first = 0;
	static struct s_hashp *last;
	struct s_hashp *hp;

	if ((hp = (struct s_hashp *) malloc(sizeof(struct s_hashp))) == NULL) {
		fprintf(stderr, "add_id: can't malloc for hashp, num_ids=%d\n", num_ids);
		exit(-1);
	}
	hp->hash = p;
	hp->next = (struct s_hashp *) NULL;
	if (first == 0) {
		first = 1;
		star_list = hp;
		last = star_list;
	}
	else {	
		last->next = hp;
		last = last->next;
	}
}

	/***********************************************************************
	 * PROCEDURE: init_hash
	 *
	 * DESCRIPTION: Initialize the hash table we use for assigning 
	 *              IDs to individual measurements.  
	 *
	 *              It's probably better to read the maximum size
	 *              of (x,y) coord system from command line than to
	 *              handle as #define, but change later.
	 *
	 * RETURNS:
	 *   nothing
	 */

static void
init_hash
	(
	void
	)
{
	int i;

	nhash = MAX_COORD;

	/* allocate an array large enough to fit 'rowmax' rows */
	if ((hashtab = (struct s_hash *) calloc(nhash, sizeof(struct s_hash))) == NULL) {
		fprintf(stderr, "init_hash: can't calloc for %d hash entries \n", 
							nhash);
		exit(-1);
	}

	/* initialize the values in the table */
	for (i = 0; i < nhash; i++) {
		hashtab[i].xc = -1.0;
		hashtab[i].yc = -1.0;
		hashtab[i].id = -1;
		hashtab[i].num = 0;
		hashtab[i].starp = (struct s_starp *) NULL;
		hashtab[i].next = (struct s_hash *) NULL;
		hashtab[i].include = 0;
	}
}

	/**************************************************************
	 * PROCEDURE: is_a_match
	 *
	 * DESCRIPTION: Compare the distance between the star 
	 *              and the hash-entry-centroid to the critical distance 
	 *              for a match; if the distance is less, 
	 *              we have a match.
	 *
	 * RETURNS:
	 *     0        if this star doesn't match the given hash entry
	 *     1        if this is a match
	 */


static int
is_a_match
	(
	struct s_star *s1,
	struct s_hash *p
	)
{
	double dr, dc;

	dr = (s1->xc - p->xc);
	dc = (s1->yc - p->yc);
	if ((dr*dr + dc*dc) <= RAD2) {
		return(1);
	}
	else {
		return(0);
	}
} 

	/**************************************************************
	 * PROCEDURE: print_hash
	 *
	 * DESCRIPTION: Print out the entire hash table - 
	 *              for debugging purposes only
	 *
	 * RETURNS:
	 *     nothing
	 */


#ifdef DEBUG
static void
print_hash
	(
	void
	)
{
	int i, n, nn;
	struct s_hash *p;
	struct s_starp *s;

	for (i = 0; i < nhash; i++) {
		if (hashtab[i].id == -1) {
			printf("hashtab[%4d]   ", i);
			printf("empty \n");
		}
		else {
			n = 0;
			printf("hashtab[%4d]   ", i);
			p = &(hashtab[i]);
			while (p != NULL) {
				if (n == 0) {
					printf("entry %3d: ID=%5d  num=%3d  xc=%6.2lf yc=%6.2lf inc=%d\n", 
							n, p->id, p->num, p->xc, p->yc, p->include);
				}
				else {
					printf("%15s entry %3d: ID=%5d  num=%3d  xc=%6.2lf yc=%6.2lf inc=%d\n", 
							" ", n, p->id, p->num, p->xc, p->yc, p->include);
				}
				nn = 0;
				s = p->starp;
				while (s != NULL) {
					printf("%20s starp=%d  xc=%6.2lf yc=%6.2lf reqid=%s\n",
								" ", nn, s->star->xc, s->star->yc, 
								s->star->image->name);
					nn++;
					s = s->next;
				}
				n++;
				p = p->next;
			}
		}
	}
}


	/**************************************************************
	 * PROCEDURE: print_by_id
	 *
	 * DESCRIPTION: Print out a nice list of stars by their ID numbers 
	 *
	 * RETURNS:
	 *     nothing
	 */

static void
print_by_id
	(
	void
	)
{
	int nn;
	struct s_hashp *hp;
	struct s_hash *p;
	struct s_starp *s;

	for (hp = star_list; hp != NULL; hp = hp->next) {
		p = hp->hash;
		printf("ID=%5d  num=%3d  xc=%6.2lf yc=%6.2lf inc=%d\n", 
						p->id, p->num, p->xc, p->yc, p->include);
		nn = 0;
		s = p->starp;
		while (s != NULL) {
			printf("%9s starp=%d  xc=%6.2lf yc=%6.2lf w3=%d w4=%-8.1lf  reqid=%s\n", 
						" ", nn, s->star->xc, s->star->yc, s->star->w3,
						s->star->w4, s->star->image->name);
			printf(" \n");
			nn++;
			s = s->next;
		}
	}

}

#endif


	/**************************************************************
	 * PROCEDURE: print_star_hist
	 *
	 * DESCRIPTION: Print out a histogram of the number of 
	 *              images in which each star appears.  The user can
	 *              use this to decide which stars ought to be included
	 *              in the solution.
	 *
	 * RETURNS:
	 *     nothing
	 */

void
print_star_hist()
{
	int i, div, hist[HIST_BINS], max;
	struct s_hashp *hp;
	struct s_hash *p;

	for (i = 0; i < HIST_BINS; i++)
		hist[i] = 0;

	max = 0;
	for (hp = star_list; hp != NULL; hp = hp->next) {
		p = hp->hash;
		if (p->num > max) {
			max = p->num;
		}
	}
	div = ((max - 1)/HIST_BINS) + 1;
	for (hp = star_list; hp != NULL; hp = hp->next) {
		p = hp->hash;
		hist[(p->num/div)]++;
	}
	for (i = 0; i < HIST_BINS; i++) {
		printf("%5d  %3d\n", (int) ((i+0.5)*div), hist[i]);
	}

}


	/**************************************************************
	 * PROCEDURE: print_image_hist
	 *
	 * DESCRIPTION: Print out a histogram of the number of images 
	 *              with at least N stars, vs. N 
	 *
	 * RETURNS:
	 *     nothing
	 */

void
print_image_hist()
{
	int i, div, hist[HIST_BINS], max;

	for (i = 0; i < HIST_BINS; i++)
		hist[i] = 0;

	max = 0;
	for (i = 0; i < num_images; i++) {
		if (image_arr[i].num_star > max) {
			max = image_arr[i].num_star;
		}
	}
	div = ((max - 1)/HIST_BINS) + 1;
	for (i = 0; i < num_images; i++) {
		hist[(image_arr[i].num_star/div)]++;
	}
	for (i = 0; i < HIST_BINS; i++) {
		printf("%5d  %3d\n", (int) ((i+0.5)*div), hist[i]);
	}

}


	/**********************************************************************
	 * PROCEDURE: get_star_index
	 *
	 * DESCRIPTION: given a STAR structure, use its "xc" value to
	 *              determine an "index" into the hash table.
	 *              We simply set
	 *                     index = xc
	 *              except that we make sure the index is within
	 *              legal ranges (i.e. >=0 and < nhash).
	 *
	 * RETURNS:
	 *     a hash index
	 *     0                   if we're handed NULL pointer
	 */

static int 
get_star_index
	(
	struct s_star *starp          /* I: pointer to STAR structure */
	)
{
	int index;

	if (starp == NULL) {
		fprintf(stderr, "get_star_index: handed NULL starp?  Return 0\n");
		return(0);
	}
	index = floor(starp->xc);
	if (index < 0) {
		index = 0;
	} 
	if (index >= nhash) {
		index = nhash - 1;
	}

	return(index);

}
