
/*
 * Modified the "write_one_star" function so that it prints to the
 *   output file sp->star->xc and sp->star->yc, without modification.
 *   The old version was using
 *          sp->star->xc - imagep->xshift
 *          sp->star->yc - imagep->yshift
 *   but that was giving output positions which wandered around;
 *   the reason is that the sp->star->xc and ->yc positions have
 *   _already_ been corrected for image shifts, so that they provide
 *   registered positions as-is when we reach "write_one_star()".
 * MWR 8/22/2004
 *
 * Added "good_stars_per_image()" function, which counts the number
 *   of stars in each image which will be included in the ensemble.
 *   The "write_one_image()" function now places _this_ value into
 *   the output file, instead of the total number of stars in each image.
 * MWR 11/20/2004
 */


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "multipht.h"

struct s_imsort {
	int index;			/* index into the image_arr[] array */
	double jd;			/* jd of index'th image */
};

struct s_stsort {
	struct s_hash *hash;	/* pointer to hash-list entry for this star */
	double xc;				/* xc coord of this star in master list */
};

static void eliminate(int flags);
static void write_it(FILE *fp);
static int good_images(void);
static int good_stars(void);
static void good_stars_per_image(void);
static void write_stars(FILE *fp);
static void write_one_star(FILE *fp, struct s_hash *p);
static int comp_stsort(struct s_stsort *s1, struct s_stsort *s2);
static void write_images(FILE *fp);
static void write_one_image(FILE *fp, struct s_image *im);
static int comp_imsort(struct s_imsort *i1, struct s_imsort *i2);


	/**************************************************************
	 * PROCEDURE: write_output
	 *
	 * DESCRIPTION: Write information on all 'acceptable' images 
	 *              and stars to the given file 
	 *              (or stdout, if passed an empty string "").
	 *
	 * RETURNS:
	 *     nothing
	 */

void
write_output
	(
	char *file, 
	int flags
	)
{
	FILE *fp;

	if (strcmp(file, "") == 0)
		fp = stdout;
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s for writing\n", file);
		exit(-1);
	}

	eliminate(flags);
	write_it(fp);

	if (fp != stdout)
		fclose(fp);

}

	/**************************************************************
	 * PROCEDURE: eliminate
	 *
	 * DESCRIPTION: Set to zero the 'w3' weight of 
	 *              all stars which are either in an 
	 *              'unacceptable' image, or an instance of an 
	 *              'unacceptable' star.  
	 *
	 *              If not "flags" is not AUTO, ask the user if 
	 *              she wishes to get rid of all saturated stars. 
	 *
	 * RETURNS:
	 *     nothing
	 */


static void 
eliminate
	(
	int flags
	)
{
	int i, j;
	char str[STRLEN];
	struct s_starp *sp;
	struct s_hashp *hp;

	for (i = 0; i < num_images; i++)
		if (image_arr[i].include == 0)
			for (j = 0; j < image_arr[i].num_star; j++)
				image_arr[i].stars[j].w3 = 0;

	for (hp = star_list; hp != NULL; hp = hp->next) 
		if (hp->hash->include == 0)
			for (sp = hp->hash->starp; sp != NULL; sp = sp->next)
				sp->star->w3 = 0;

	if ((flags & AUTO) == 0) {
		printf("do you wish to eliminate all saturated stars (y/n)? ");
		scanf("%s", str);
		if ((str[0] == 'y') || (str[0] == 'Y')) {
			for (i = 0; i < num_images; i++)
				for (j = 0; j < image_arr[i].num_star; j++)
					if (image_arr[i].stars[j].saturated != 0)
						image_arr[i].stars[j].w3 = 0;
		}
	}
		
}


	/**************************************************************
	 * PROCEDURE: write_it
	 *
	 * DESCRIPTION: Write output for images, first, and then stars.
	 *
	 * RETURNS:
	 *     nothing
	 */

static void 
write_it
	(
	FILE *fp
	)
{
	int goodim, goodstar;

	/* there are two kinds of information to print out: information 
	   about images (in the 'image_arr[]' array), and information
	   about stars (in the 'star_list' list of hashp structures. 
	   the first thing to do is to write how many images and how many
	   stars will appear in the output. */

	goodim = good_images();
	goodstar = good_stars();
	fprintf(fp, "%4d images  %6d stars \n", goodim, goodstar);

	/* count the number of stars we'll include from each image */
	good_stars_per_image();

	/* now write information on each image: use one line per image */
	write_images(fp);	

	/* and now write information on each star: use N+1 lines per star,
	   where N is the number of 'acceptable' images of each star */
	write_stars(fp);

}

	/**************************************************************
	 * PROCEDURE: good_images
	 *
	 * DESCRIPTION: Calculate the number of 'good' images:
	 *              the number which will be included in the solution.
	 *
	 * RETURNS:
	 *     the number of good images
	 */


static int 
good_images
	(
	void
	)
{
	int i, sum;

	sum = 0;
	for (i = 0; i < num_images; i++) {
		if (image_arr[i].include != 0)  {
			sum++;
		}
	}
	return(sum);
}

	/**************************************************************
	 * PROCEDURE: good_stars
	 *
	 * DESCRIPTION: Calculate the number of 'good' stars:
	 *              the number of stars which will be included
	 *              in the solution.
	 *
	 *              There must be at least one instance
	 *              of each good star; otherwise, it isn't good.
	 * RETURNS:
	 *     the number of stars in solution
	 */

static int
good_stars
	(
	void
	)
{
	int sum;
	struct s_starp *sp;
	struct s_hashp *hp;

	sum = 0;
	for (hp = star_list; hp != NULL; hp = hp->next) {
		if (hp->hash->include != 0) {
			for (sp = hp->hash->starp; sp != NULL; sp = sp->next) {
				if (sp->star->w3 != 0) {
					sum++;
					break;
				}
			}
		}
	}
	
	return(sum);
}




	/**************************************************************
	 * PROCEDURE: good_stars_per_image
	 *
	 * DESCRIPTION: Calculate the number of 'good' stars
	 *              in each image: 
	 *              the number which will be included in the solution.
	 *
	 *              This routine will set the 'num_good_star' field
	 *              within each image structure.
	 *
	 * RETURNS:
	 *     nothing
	 */


static void
good_stars_per_image
	(
	void
	)
{
	int i, j, sum;
	struct s_star *star;

	for (i = 0; i < num_images; i++) {
		if (image_arr[i].include != 0)  {
			sum = 0;
			for (j = 0; j < image_arr[i].num_star; j++) {
				star = &(image_arr[i].stars[j]);
				if (star->w3 != 0) {
					sum++;
				}
			}
			image_arr[i].num_good_star = sum;	
		}
	}
}



	/**************************************************************
	 * PROCEDURE: write_stars
	 *
	 * DESCRIPTION: Write out information on each star, 
	 *              putting them in order by master list x-coordinate.
	 *
	 * RETURNS:
	 *     nothing
	 */


static void
write_stars
	(
	FILE *fp
	)
{
	int i;
	struct s_stsort *stsort;
	struct s_hash *hash;
	struct s_hashp *hp;

	/* effectively, sort the stars by their master xc coord */
	if ((stsort = (struct s_stsort *) calloc(num_ids, sizeof(struct s_stsort))) == NULL) {
		fprintf(stderr, "write_stars: can't calloc for stsort array \n");
		exit(-1);
	}
	for (i = 0, hp = star_list; (i < num_ids) && (hp != NULL); i++, hp = hp->next) {
		stsort[i].hash = hp->hash;
		stsort[i].xc = hp->hash->xc;
	}	
	qsort((char *) stsort, num_ids, sizeof(struct s_stsort), (PFI) comp_stsort);

	for (i = 0; i < num_ids; i++) {
		hash = stsort[i].hash;
		if (hash->include != 0)
			write_one_star(fp, hash);
	}
}


	/**************************************************************
	 * PROCEDURE: write_one_star
	 *
	 * DESCRIPTION: Write the pertinent information from 
	 *              a single hash-entry structure 
	 *              (which records all instances of the same star) 
	 *              to the given FILE pointer.  Each 'master star' 
	 *              gets N+1 lines of output, where N is the number 
	 *              of 'good' instances of the star.  
	 *              The first line includes the number N at its end,
	 *              so programs which read the file know how many
	 *              to expect.
	 *
	 * RETURNS:
	 *     nothing
	 */


static void
write_one_star
	(
	FILE *fp, 
	struct s_hash *hash
	)
{
	int good;
	static int newid = 0;
	struct s_starp *sp;
	struct s_image *imagep;
	double xpos, ypos;

	good = 0;
	for (sp = hash->starp; sp != NULL; sp = sp->next) {
		if (sp->star->w3 != 0) {
			good++;
		}
	}

	fprintf(fp, "%6d %7.2lf %7.2lf %6d \n", newid, hash->xc, hash->yc, good);
	for (sp = hash->starp; sp != NULL; sp = sp->next) {
		if (sp->star->w3 != 0) {
			/* calculate original position of this star on its image */
			imagep = &(image_arr[sp->star->image->index]);
#if 0
			/* 
			 * this yields output (x, y) positions which wander from 
			 *   image to image, and don't _quite_ match the actual
			 *   original values in the input files.
			 *   so let's not use them ....
			 */
			xpos = sp->star->xc - imagep->xshift;
			ypos = sp->star->yc - imagep->yshift;
#else
			/* 
			 * instead, let's print out the positions after they have 
			 *   been shifted to match the template image.
			 */
			xpos = sp->star->xc;
			ypos = sp->star->yc;
#endif

			fprintf(fp, "  %6d %6d %12.5f %12.5f %1d  %7.4lf %7.1lf\n",
						newid, sp->star->image->index, 
						xpos, ypos, sp->star->saturated,
						sp->star->mag, sp->star->w4);
		}
	}
	newid++;
}


	/**************************************************************
	 * PROCEDURE: comp_stsort
	 *
	 * DESCRIPTION: Compare the "xc" values for the two stars.
	 *
	 * RETURNS:
	 *       -1 if the first star has an xc smaller than the second
	 *        0                             equal to 
    *       +1                             larger than
	 */

static int
comp_stsort
	(
	struct s_stsort *s1, 
	struct s_stsort *s2
	)
{
	if (s1->xc < s2->xc) {
		return(-1);
	}
	if (s1->xc > s2->xc) {
		return(1);
	}
	return(0);
}
		

	/**************************************************************
	 * PROCEDURE: write_images
	 *
	 * DESCRIPTION: Write out information on each image, 
	 *              putting them in order by date and time 
	 *              of observation.
	 *
	 * RETURNS:
	 *     nothing
	 */


static void
write_images
	(
	FILE *fp
	)
{
	int i, index, n;
	struct s_imsort *imsort;

	/* effectively, sort the images by their time of observation */
	if ((imsort = (struct s_imsort *) calloc(num_images, sizeof(struct s_imsort))) == NULL) {
		fprintf(stderr, "write_images: can't calloc for imsort array \n");
		exit(-1);
	}
	for (i = 0; i < num_images; i++) {
		imsort[i].index = i;
		imsort[i].jd = image_arr[i].jd;
	}	
	qsort((char *) imsort, num_images, sizeof(struct s_imsort), (PFI) comp_imsort);

	n = 0;
	for (i = 0; i < num_images; i++) {
		index = imsort[i].index;
		if (image_arr[index].include != 0) {
			image_arr[index].index = n++;
			write_one_image(fp, &(image_arr[index]));
		}
	}
}



	/**************************************************************
	 * PROCEDURE: good_images
	 *
	 * DESCRIPTION: Write the pertinent information from a 
	 *              single image structure to the given FILE pointer.
	 *
	 * RETURNS:
	 *     nothing
	 */

static void
write_one_image
	(
	FILE *fp,
	struct s_image *im
	)
{

	fprintf(fp, "%6d %30s %6.1lf %12.5lf %6d %6d\n",
			im->index, im->name, im->exptime, 
			im->jd, im->num_good_star, im->num_match);

}



	/**************************************************************
	 * PROCEDURE: comp_imsort
	 *
	 * DESCRIPTION: Compare the "JD" values of two images.
	 *
	 * RETURNS:
	 *      -1 if the first image has an jd before the second
	 *       0                               same as
    *      +1                               after
	 */

static int
comp_imsort
	(
	struct s_imsort *i1, 
	struct s_imsort *i2
	)
{
	if (i1->jd < i2->jd) {
		return(-1);
	}
	if (i1->jd > i2->jd) {
		return(1);
	}
	return(0);
}
		

