

	/*********************************************************************
	 * These routines set up a series of simultaneous equations which 
	 * correspond to the photometric solution of a large number
	 * of observations.  We set up a big matrix equation of the form
	 *
	 *              A  *  x  =   b
	 *
	 * where 'A' is a matrix, 'b' is a vector with the observed values of
	 * instr. mags and weights, and 'x' is the unknown vector of
	 * magnitude adjustments that minimizes the sums of the squares
	 * of the residuals from the mean. 
	 *
	 * Then, we call some Numerical Recipes code is used to solve it.
	 *
	 * MWR 10/18/2003 
	 *
	 *
	 * Added FORCE_STAR_MAG option, which may help solution under
	 *   some circumstances.
	 * Added a shift to the image adjustment values so that the
	 *   input (raw) mags and output (true) mags do obey the relationship
	 *          true_mag   +   image_adj   =   raw_mag
	 *
	 * MWR 10/28/2003
	 *
	 * Modified code to include terms for zero-point offsets of each
	 * filterchip.  That means that we now solve for magnitude of each
	 * star, adjustment of each image, and offset of each chip.
	 *
	 * MWR 3/23/2004
	 *
	 * Fixed huge error in "zero_vector()" which actually prevented
	 * any elements from being set to zero (!).
	 * MWR 3/25/2004
	 *
	 * Commented out the "calc_model_mag()" function, since it  
	 * isn't used any more in this simplified version.
	 * MWR 11/8/2004
	 *
	 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "solvepht.h"
#include "nr.h"

#define TOLERANCE    1.0e-7
#define TINY         (2.0*TOLERANCE) 
#define MAG_MATCH    0.30		/* used by find_sigm() */

#undef DEBUG


  /* 
	* if this is defined, then we set matrix elements corresponding 
	*   to the first star in the list to be zero.   That means that
	*   the first star's mag in the output should be zero, too.
	*   After solving the matrix equation, we need to go back and 
	*   set the output vector x element for this first star to be
	*   zero.  And _then_ we can run the normalize() function,
	*   which will shift the output magnitude values so that the
	*   BRIGHTEST star has average mag zero, rather than the first
	*   star in the list.
	*
	* Setting this option may cause the matrix solution to behave
	*   better.  In some cases, however, the code doesn't need us
	*   to take this step; it returns a reasonable solution on its
	*   own.
	*/
#define FORCE_STAR_MAG 



static void init_matrix(int ee, int ss);
static void calc_em(int m, int n);
static void calc_mags(int m, int n);
static void calc_sigmas(int m, int n);
static void zero_vector(double *p, int m, double tolerance);
static void print_vector(double *p, int m);
static void print_matrix(double **p, int m, int n);
static void fill_bvec(double *b, int m, int n);
static void read_data(int ee, int ss);
static void fill_matrix(double **p, int ee, int ss);
static void copy_matrix(double **p, double **q, int m, int n);
static void normalize(int m, int n, double value);
static void find_means(int m, int n);
static double find_sigm(int n, int e, int s);
static void print_curves(int m, int n);
static void print_sigmas(int m, int n);
static double calc_model_correction(int num_images, int num_stars, 
		                        int image_index, int star_index);
#if 0
static double calc_model_mag(int num_images, int num_stars, 
		                        int image_index, int star_index);
#endif


	/* these are inputs */
static double **mag;          /* ee by ss matrix of instr. mags */
static double **weight;			/* ee by ss matrix of weights */

	/* these are computational values */
static double **a;				/* ee+ss+cc by ee+ss+cc matrix; A above */
static double **u;				/* ee+ss+cc by ee+ss+cc: part 1 of the SVD output */
static double *w;				/* ee+ss+cc vector: part 2 of SVD output */
static double **v;				/* ee+ss+cc by ee+ss+cc: part 3 of SVD output */
static double *b;				/* ee+ss+cc vector: 'b' above */
static double *x;				/* ee+ss+cc vector: 'x' above, the 'solution' */
	/* and these are output results */
static double *em;				/* ee vector: mean correction for image e */
static double *mag0;				/* ss vector: mean mag of star s */
static double *sigm0;			/* ss vector: sigma of mean value of star s */
static double **Mes;				/* ee by ss matrix: solved mag of star s */
                   				/*    on image e */
static double **sigMes;			/* ee by ss matrix: total sigma of */
                  				/*    solution's value of star s on image e */
static double *sigem;			/* ee vector: sigma of exp corr. of image e */
static double *sigemn;			/* ee vector: sigem/# stars in image e */



	/*********************************************************************
	 * PROCEDURE: solve_matrix
	 *
	 * DESCRIPTION: set up all the equations and solve them.
	 *              If the debug flag is set, print out some results.
	 *
	 * RETURNS:
	 *    nothing
	 */

void
solve_matrix
	(
	double zero_point       /* I: set brightest star to this mag on output */
	)
{
	int ee, ss;
	double tolerance;

	tolerance = TOLERANCE;
	ee = num_images;
	ss = num_stars;
	init_matrix(ee, ss);

	read_data(ee, ss);
	fill_matrix(a, ee, ss);
	if (debug_flag) {
		printf("\n This is the matrix \n");
		print_matrix(a, ee + ss, ee + ss);
	}
	copy_matrix(a, u, ee + ss, ee + ss);

	fill_bvec(b, ee, ss);
	if (debug_flag) {
		printf(" This is the B vector \n");
		print_vector(b, ee + ss);
	}

	svdcmp(u, ee + ss, ee + ss, w, v);

	if (debug_flag) {
		printf("this is the svd matrix U \n");
		print_matrix(u, ee + ss, ee + ss);
		printf("this is the vector of w(j) - which could need zero'ing \n");
		print_vector(w, ee + ss);
	}

	zero_vector(w, ee + ss, tolerance);
	if (debug_flag) {
		printf("this is the vector w(j) after zeroing tiny elements\n");
		print_vector(w, ee + ss);
	}

	svbksb(u, w, v, ee + ss, ee + ss, b, x);

	if (debug_flag) {
		printf("\nand this is the vector of x - the solution\n");
		print_vector(x, ee + ss);
	}

#ifdef FORCE_STAR_MAG
	/* 
	 * the first star's mag in the solution vector x has been unconstrained
	 *   if we chose the FORCE_STAR_MAG option.  We need to set it equal
	 *   to zero explicitly before we go any further.
	 */
	x[ee + 1] = 0.0;
#endif

	calc_em(ee, ss);
	calc_mags(ee, ss);
	find_means(ee, ss);
	normalize(ee, ss, zero_point);
	calc_sigmas(ee, ss);

	if (debug_flag) {
		print_curves(ee, ss);
		print_sigmas(ee, ss);
	}
}


	/************************************************************************
	 * PROCEDURE: init_matrix
	 *
	 * DESCRIPTION: allocate space for all the 1-indexed arrays and matrices
	 *              that will be needed to set up and solve the series
	 *              of linear equations.
	 *
	 * RETURNS:
	 *    nothing
	 */


static void
init_matrix
	(
	int ee,              /* I: number of images in solution */
	int ss               /* I: number of stars in solution */
	)
{

	if ((mag = matrix(1, ee, 1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix mag\n");
		exit(-1);
	}
	if ((weight = matrix(1, ee, 1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix weight\n");
		exit(-1);
	}
	if ((a = matrix(1, ee + ss, 1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix a\n");
		exit(-1);
	}
	if ((b = vector(1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector b \n");
		exit(-1);
	}
	if ((u = matrix(1, ee + ss, 1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix u\n");
		exit(-1);
	}
	if ((v = matrix(1, ee + ss, 1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix v\n");
		exit(-1);
	}
	if ((w = vector(1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector w \n");
		exit(-1);
	}
	if ((x = vector(1, ee + ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector x \n");
		exit(-1);
	}
	if ((em = vector(1, ee)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector em \n");
		exit(-1);
	}
	if ((mag0 = vector(1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector mag0 \n");
		exit(-1);
	}
	if ((sigm0 = vector(1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigm0 \n");
		exit(-1);
	}
	if ((Mes = matrix(1, ee, 1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix Mes \n");
		exit(-1);
	}
	if ((sigMes = matrix(1, ee, 1, ss)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for matrix sigMes \n");
		exit(-1);
	}
	if ((sigem = vector(1, ee)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigem \n");
		exit(-1);
	}
	if ((sigemn = vector(1, ee)) == NULL) {
		fprintf(stderr, "init_matrix: can't allocate for vector sigemn \n");
		exit(-1);
	}
}



	/**********************************************************************
	 * PROCEDURE: calc_sigmas
	 *
	 * DESCRIPTION: After we have solved the equations, we can go back
	 *              and calculate values of the RMS for each exposure 
	 *              (as a whole), for mean magnitude of each star, 
	 *              and for each observation of each star (individually).
	 *
	 *              The tricky part is to handle values of NODATA 
	 *              correctly so that they don't throw averages off. 
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
calc_sigmas
	(
	int m,              /* I: number of images in solution */
	int n               /* I: number of stars in solution */
	)
{
	int i, j, num;
	double temp, top, bot;
	struct s_measurep *mp;
	struct s_measure *measure;

	/* set everything to 0.0, first of all */
	for (i = 1; i <= m; i++) {
		sigem[i] = 0.0;
		sigemn[i] = 0.0;
		for (j = 1; j <= n; j++) {
			sigMes[i][j] = 0.0;
		}
	}
	for (i = 1; i <= n; i++) {
		sigm0[i] = 0.0;
	}

	/* first find sigma(em(e)) */
	if (n >= 2) {
		for (i = 1; i <= m; i++) {
			num = 0;
/****
			for (sp = image_arr[i - 1].starp; sp != NULL; sp = sp->next)
				if (sp->star->include != 0)
					num++;
***/
			top = 0.0;
			bot = 0.0;
			for (j = 1; j <= n; j++) {
				if ((Mes[i][j] == NODATA) || (mag0[j] == NODATA)) {
					continue;
				}
				num++;
				temp = (Mes[i][j] - mag0[j]);
				top += temp*temp*weight[i][j];
				bot += weight[i][j];
			}
			if (bot > 0.0) {
				sigem[i] = sqrt((n*top)/((n - 1)*bot));
				if (num > 0) {
					sigemn[i] = sigem[i]/(double) sqrt((double) num);
				}
			}
			else {
				sigem[i] = NODATA;	
				sigemn[i] = NODATA;	
			}
		}
	}

	/* now, find sigma(m0(s)) */
	if (m >= 2) {
		for (i = 1; i <= n; i++) {
			top = 0.0;
			bot = 0.0;
			if (star_arr[i - 1].variable == 0) {
				for (j = 1; j <= m; j++) {
					if ((Mes[j][i] == NODATA) || (mag0[i] == NODATA) ||
										(image_arr[j - 1].include == 0)) {
						continue;
					}
					temp = (Mes[j][i] - mag0[i]);
					top += temp*temp*weight[j][i];
					bot += weight[j][i];
				}
			}
			else {
				for (mp = star_arr[i - 1].mp; mp != NULL; mp = mp->next) {
					measure = mp->measure;
					shAssert(measure != NULL);
					j = measure->image_id + 1;
					if ((Mes[j][i] == NODATA) || (mag0[i] == NODATA) ||
										(image_arr[j - 1].include == 0)) {
						continue;
					}
					temp = (Mes[j][i] - mag0[i]);
					top += temp*temp*measure->w4;
					bot += measure->w4;
				}
			}
			if (bot > 0.0) {
				sigm0[i] = sqrt((m*top)/((m - 1)*bot));
			}
			else {
				sigm0[i] = NODATA;
			}
		}
	}

	/* now we can finally calculate sigma(M(e,s)) */
	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			temp = find_sigm(n, i, j);
			if (temp == NODATA) {
				sigMes[i][j] = NODATA;
			}
			else if (sigemn[i] != NODATA) {
				sigMes[i][j] = sqrt(temp*temp + sigemn[i]*sigemn[i]);
			}
			else {
				sigMes[i][j] = temp;
			}
		}
	}
}


	/************************************************************************
	 * PROCEDURE: print_sigmas
	 *
	 * DESCRIPTION: print out some derived quantities.  For debugging
	 *              purposes only.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
print_sigmas
	(
	int m,            /* I: number of images */
	int n             /* I: number of stars */
	)
{
	int i;

	/* first find sigma(em(e)) */
	printf("   Sigma of Exposures: image #, sigem, sigemn\n");
	for (i = 1; i <= m; i++) {
		printf(" %4d  %5.3f  %5.3f\n", image_arr[i - 1].index, sigem[i], sigemn[i]); 
	}
	printf("\n");

	/* now, find sigma(m0(s)) */
	printf("  Sigma of stellar mean mags: star #, m0, sig(m0) \n");
	for (i = 1; i <= n; i++) {
		printf("  %6d %6.3f %5.3f  \n", star_arr[i - 1].id, mag0[i], sigm0[i]); 
	}
	printf("\n");
}



	/***********************************************************************
	 * PROCEDURE: calc_em
	 *
	 * DESCRIPTION: Calculate the 'em' matrix - the overal shift of 
	 *              each image's gnitudes to bring them close to the mean 
	 *              values. 
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
calc_em
	(
	int m,          /* I: number of images */
	int n           /* I: number of stars */
	)
{
	int i;
	
	for (i = 1; i <= m; i++) { 
		if (image_arr[i - 1].include) {
			em[i] = x[i];
		}
		else {
			em[i] = NODATA;
		}
	}
}


	/***********************************************************************
	 * PROCEDURE: calc_mags
	 *
	 * DESCRIPTION: Adjust the instrumental magnitudes for each star, 
	 *              adding the results of the matrix inversion to the
	 *              raw instrumental magnitude.  The result is the
	 *              differential magnitude of the star relative to
	 *              all the others in all images.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
calc_mags
	(
	int m,          /* I: number of images */
	int n           /* I: number of stars */
	)
{
	int i, j;	
	double model_correction;
	
	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			if ((mag[i][j] != NODATA) && (image_arr[i - 1].include != 0)) {
				model_correction = calc_model_correction(m, n, i, j);
#ifdef DEBUG
				printf ("  calc_mags: %6.3lf %6.3lf   %6.3lf \n",
						mag[i][j], model_correction, mag[i][j] - model_correction);
#endif
				if (model_correction != NODATA) {
					Mes[i][j] = mag[i][j] - model_correction;
				}
				else {
					Mes[i][j] = NODATA;
				}
			}
			else {
				Mes[i][j] = NODATA;
			}
		}
	}
}





	/************************************************************************
	 * PROCEDURE: print_curves
	 *
	 * DESCRIPTION: Print out light curves for each star 
	 *              (used only in debugging)
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
print_curves
	(
	int m,             /* I: number of images */
	int n              /* I: number of stars */
	)
{
	int i, j;	
	
	printf("\n");
	for (i = 1; i <= n; i++) {
		printf("star   %2d: ", i);
		for (j = 1; j <= m; j++) {
			printf("%6.3f %6.3f  ", Mes[j][i], sigMes[j][i]);
		}
		printf("\n");
	}
}




	/************************************************************************
	 * PROCEDURE: zero_vector
	 *
	 * DESCRIPTION: Set to zero all elements of vector p which are 
	 *              less than 'tolerance' times the maximum element of p
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
zero_vector
	(
	double *p,            /* I/O: a vector whose elements we may set to zero */
	int m,                /* I: number of elements in vector */
	double tolerance      /* I: if elem i < this times max elem, set to 0  */
	)
{
	int i;
	double max, min;

	max = 0.0;
	for (i = 1; i <= m; i++) {
		if (p[i] > max) {
			max = p[i];
		}
	}
	min = tolerance*max;
	for (i = 1; i <= m; i++) {
		if (p[i] < min) {
			p[i] = 0.0;
		}
	}
}	



	/*******************************************************************
	 * PROCEDURE: print_vector
	 *
	 * DESCRIPTION: Print a vector (which goes 1-m) in a nice format
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
print_vector
	(
	double *p,          /* I: the vector we'll print */
   int m               /* I: number of elements in the vector */
	)
{
	int i;

	for (i = 1; i <= m; i++) {
		printf("%9.3e ", p[i]);
	}
	putchar('\n');
}


	/**********************************************************************
	 * PROCEDURE: print_matrix
	 *
	 * DESCRIPTION: Print a matrix (which goes 1-m, 1-n) in a nice format 
	 *
	 * RETURNS:
	 *   nothing
	 */

static void
print_matrix
	(
	double **p,            /* I: matrix whose elements we print */
	int m,                 /* I: number of rows in matrix */
	int n                  /* I: number of columns in vector */
	)
{
	int i, j;

	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			printf("%9.3e ", p[i][j]);
		}
		putchar('\n');
	}
}



	/***********************************************************************
	 * PROCEDURE: fill_bvec
	 *
	 * DESCRIPTION: Fill up the vector B, which is computed from 
	 *              observed magnitudes and weight.  This sits on the
	 *              right-hand side of the matrix equation.
	 *
	 *              There are three pieces to the B vector.  
	 *                 First, a set of ee values, 1 per image
	 *                 Next,  a set of ss values, 1 per star
	 *
	 * RETURNS:
	 *    nothing
	 */

static void 
fill_bvec
	(
	double *b,             /* O: fill elements of this vector */
	int m,                 /* I: number of images */
	int n                  /* I: number of stars */
	)
{
	int i, j;
	double x;

	/* 
	 * add up weights in each image; 
	 *   this includes an implicit sum over all filterchips
	 */
	for (i = 1; i <= m; i++) {
		x = 0;
		for (j = 1; j <= n; j++) {
			x += weight[i][j]*mag[i][j];
		}
		b[i] = x;
	}

	/* 
	 * add up weights for each star; 
	 *   this includes an implicit sum over all filterchips
	 */
	for (j = 1; j <= n; j++) {
		x = 0;
		for (i = 1; i <= m; i++) {
			x += weight[i][j]*mag[i][j];
		}
		b[j + m] = x;
	}

}


	/************************************************************************
	 * PROCEDURE: read_data
	 *
	 * DESCRIPTION: take the magnitude and uncertainty data, and figure
	 *              out how to turn it into weights for the
	 *              linear equations.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
read_data
	(
	int m,              /* I: number of images */
	int n              /* I: number of stars */
	)
{
	int i, j;
	int num_measures;
	struct s_star *star;
	struct s_measurep *mp;
	struct s_measure *measure;

	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			mag[i][j] = 0.0;
			weight[i][j] = 0.0;
		}
	}

	for (j = 1; j <= n; j++) {
		star = &(star_arr[j - 1]);	
		shAssert(star != NULL);
		mp = star->mp;
		num_measures = 0;
		while (mp != NULL) {
			measure = mp->measure;
			shAssert(measure != NULL);
			i = measure->image_id + 1;
			mag[i][j] = (double) measure->obs_mag;
			if ((star->include != 0) && (star->variable == 0) &&
			    (measure->include != 0) &&
				 (measure->image->include != 0)) {
				weight[i][j] = (double) measure->w4;
			}
#ifdef DEBUG
			printf("  j %6d  i %6d  mag %6.3f  weight %6.3f \n", j, i, 
					mag[i][j], weight[i][j]);
#endif

			num_measures++;
			mp = mp->next;
		}
		if (num_measures != star->num_measures) {
			fprintf(stderr, "read_data: num_measures %d but star_num %d ?!",
					num_measures, star->num_measures);
			exit(0);
		}
	}

}



	/***********************************************************************
	 * PROCEDURE: fill_matrix
	 *
	 * DESCRIPTION: Fill up the big matrix with properly weighted quantities.
	 *
	 *              The matrix can be broken into four "pieces"
	 *
	 *                   A      B      
	 *                   
	 *                   C      D 
	 *
	 *                   
	 *              where piece "A"  sum over all stars for each image
	 *                          "B"  weights of each image, star
	 *                          "C"  same as "B"
	 *                          "D"  sum over all images for each star
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
fill_matrix
	(
	double **p,           /* O: fill the elements of this matrix */
	int m,                /* I: number of images */
	int n                 /* I: number of stars */
	)
{
	int i, j;
	double x;

	for (i = 1; i <= m + n; i++) {
		for (j = 1; j <= m + n; j++) { 	
			p[i][j] = 0.0;
		}
	}

	/* fill up the real matrix in four steps, one per piece */

	/*   piece "A": sum weights of all stars in each image */
	for (i = 1; i <= m; i++) {
		x = 0;
		for (j = 1; j <= n; j++) {
			x += weight[i][j];
		}
		p[i][i] = x;
	}

	/*   pieces "B" and "C": weight for each individual star, image */
	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			p[i][j + m] = weight[i][j];
			p[j + m][i] = weight[i][j];
		}
	}

	/*   piece "D": sum weights for all instances of each star */
	for (i = 1; i <= n; i++) {
		x = 0;
		for (j = 1; j <= m; j++) {
			x += weight[j][i];
		}
		p[i + m][i + m] = x;
	}


#ifdef FORCE_STAR_MAG
	/* now we force the first star to have instrumental mag 0 */
	for (i = 1; i <= m + n; i++) {
		p[i][m + 1] = 0.0;
	}
#endif

}


	/************************************************************************
	 * PROCEDURE: copy_matrix
	 *
	 * DESCRIPTION: copy the values of one matrix to another one
	 *
	 * RETURNS:
	 *    nothing
	 */

static void 
copy_matrix
	(
	double **p,         /* I: copy FROM this matrix ... */
	double **q,         /* O: INTO this matrix */
	int m,              /* I: number of rows in matrices */
	int n               /* I: number of columns in matrices */
	)
{
	int i, j;

	for (i = 1; i <= m; i++) {
		for (j = 1; j <= n; j++) {
			q[i][j] = p[i][j];
		}
	}
}


	/***********************************************************************
	 * PROCEDURE: normalize
	 *
	 * DESCRIPTION: add a constant to all magnitudes 
	 *              (and to the x[] vector as well) so that the brightest 
	 *              star has the given mean magnitude value.
	 *              And also SUBTRACT this same constant to the image
	 *              adjustment factors so that we restore the 
	 *              equation connecting raw and corrected magnitudes
	 *              via the image adjustments.
	 *
	 * RETURNS:
	 *    nothing
	 *
	 */

static void
normalize
	(
	int m, 
	int n, 
	double value
	)
{
	int i, j;
	double min, diff;

	min = 1.0e6;
	for (i = 1; i <= n; i++) {
		if ((mag0[i] < min) && (mag0[i] != NODATA)) {
			min = mag0[i];
		}
	}
	diff = (double) (value - min);

	for (i = 1; i <= n; i++) {
		if (mag0[i] != NODATA) {
			mag0[i] += diff;
		}
		for (j = 1; j <= m; j++) {
			if (Mes[j][i] != NODATA) {
				Mes[j][i] += diff;
			}
		}
	}

	/* 
	 * now, shift the image adjustment magnitudes by this same amount
	 * but in the opposite direction, to preserve the connection
	 * between input and output magnitudes
	 */
	for (i = 1; i <= m; i++) { 
		if (em[i] != NODATA) {
			em[i] -= diff;
		}
	}
			
}



	/***********************************************************************
	 * PROCEDURE: find_means
	 *
	 * DESCRIPTION: find the mean magnitude of each star, either
	 *
	 *                 a. by simply looking in the 'x' vector, for 
	 *                    stars with non-zero weights
	 *                    (we solved for the mean mags of these stars)
	 *
	 *               or 
	 *
	 *                 b. by calculating it from the weighted average of
	 *                    (corrected) mag on each night 
	 *                    (these are "bad" or "variable" stars)
	 *
	 *               If there is no information for a star, 
	 *               use the NODATA value as its mean mag 
	 *
	 *               Once we know the value for a star, set its entry
	 *               in the "mag0" array to that value.
	 *
	 * RETURNS:
	 *    nothing
	 */

static void
find_means(m, n)
int m, n;
{
	int i, index;
	double top, bot;
	struct s_measurep *mp;
	struct s_measure *measure;

	for (i = 1; i <= n; i++) {
		if ((star_arr[i - 1].include != 0) && (star_arr[i - 1].variable == 0)) {
			mag0[i] = x[i + m];
		}
		else if (star_arr[i - 1].variable != 0) {
			top = 0.0;
			bot = 0.0;
			mp = star_arr[i - 1].mp;
			while (mp != NULL) {
				measure = mp->measure;
				shAssert(measure != NULL);

				index = measure->image_id + 1;
				if (Mes[index][i] == NODATA) {
					mp = mp->next;
					continue;
				}
				top += Mes[index][i]*measure->w4;
				bot += measure->w4;
				mp = mp->next;
			}

			if (bot > 0.0) {
				mag0[i] = top/bot;
			}
			else {
				mag0[i] = NODATA;
			}

		}
	}
}


	/********************************************************************
	 * PROCEDURE: find_sigm
	 *
	 * DESCRIPTION: Return an estimate of the uncertainty sigma(m(e,s)) 
	 *              for star s in image e, which is only PART of the 
	 *              total error estimate sigma(M(e,s)).  
	 *              Find this value in two ways:
	 * 
	 *                1. try to find the average value of sigma(m0) for
	 *                          m0 ~= m(e,s)
	 *
	 *                2. if impossible, then use a value calculated
	 *                   from the weight of the value for m(e,s)
	 *                   since it is really based on input estimate
	 *                   of uncertainty, like so:
 	 *                      sig(m(e,s)) = 1.0/sqrt(weight(e,s))
	 *
	 *              If neither is possible, then the uncertainty
	 *              will be NODATA
	 *
	 * RETURNS:
	 *    the value of uncertainty in mean magnitude
	 */

static double
find_sigm
	(
	int n,              /* I: number of stars overall in solution */
	int e,              /* I: we are interested in a star in this image ... */
	int s               /* I: ... and it's star 's' in this image */
	)
{
	int i, num;
	double dist, sum, wt, ret;

	/* first, try to find a mean m0 whose magnitude is close to 
	   m(e,s).  use a weighted average of the nearby sigm0. */
	sum = 0.0;
	wt = 0.0;
	num = 0;
	for (i = 1; i <= n; i++) {
		if ((dist = fabs(mag0[i] - Mes[e][s])) < MAG_MATCH) {
			if (dist < TINY) {
				dist = TINY;
			}
			sum += sigm0[i]/dist;
			wt += 1.0/dist;
			num++;
		}
	}
	if (num > 0) {
		ret = sum/wt;
		return(ret);
	}

	/* second best: use the weight of instrumental mag(e,s) as an
	   estimator */
	if (weight[e][s] > 0.0) {
		ret = 1.0/sqrt(weight[e][s]);
		return(ret);
	}

	/* nothing left to do but return NODATA */
	ret = NODATA;
	return(ret);
}



	/***********************************************************************
	 * PROCEDURE: write_images
	 *
	 * DESCRIPTION: Write output about images ONLY to the given file 
	 *              (or stdout, if the given string is "").
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 *
	 */

int
write_images
	(
	char *file
	)
{
	int i;
	FILE *fp;
	struct s_image *im;

	if (strcmp(file, "") == 0) {
		fp = stdout;
	}
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}

	for (i = 0; i < num_images; i++) {
		if (image_arr[i].include == 0) {
			continue;
		}
		im = &(image_arr[i]);
		fprintf(fp, "%-4d %30s %6.1f  %11.5f ",
			im->index, im->name, im->exptime, im->jd);
		fprintf(fp, "%5.3f %5.3f %5.3f\n", 
							em[i + 1], sigem[i + 1], sigemn[i + 1]);
	}
	if (fp != stdout) {
		fclose(fp);
	}

	return(0);
}


	/***********************************************************************
	 * PROCEDURE: write_stars
	 *
	 * DESCRIPTION: Write information about stars ONLY, one line per
	 *              star per image, to the given file 
	 *              (or stdout if file == "").  
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 *
	 */

int
write_stars(file)
char *file;
{
	int i, index;
	FILE *fp;
	struct s_star *star;
	struct s_image *im;
	struct s_measurep *mp;
	struct s_measure *measure;

	if (strcmp(file, "") == 0) {
		fp = stdout;
	}
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(0);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(0);
	}

	for (i = 0; i < num_stars; i++) {
		if (star_arr[i].include == 0) {
			continue;
		}
		star = &(star_arr[i]);
		mp = star->mp;
		while (mp != NULL) {
			measure = mp->measure;
			shAssert(mp != NULL);

			im = measure->image;
			shAssert(im != NULL);
			if (im->include != 0) {
				index = im->index;
				fprintf(fp, "%-6d %9.5f %9.5f %4d %11.5f ",
						star->id, star->ra, star->dec, index, im->jd);
				fprintf(fp, " %6.3f %6.3f %5.3f\n", mag0[i + 1], 
						Mes[index + 1][i + 1], sigMes[index + 1][i + 1]);
			}
			mp = mp->next;
		}
	}
	if (fp != stdout) {
		fclose(fp);
	}

	return(0);
}





	/************************************************************************
	 * PROCEDURE: write_sigma
	 *
	 * DESCRIPTION: Write information about sigma0 vs. mag0 
	 *              (standard deviation from mean mag, versus mean mag)
	 *              to the given file (or stdout if the given file == "") 
	 *
	 * RETURNS:
	 *    0         if all goes well
	 *    1         if an error occurs
	 *
	 */

int
write_sigma
	(
	char *file
	)
{
	int i;
	FILE *fp;

	if (strcmp(file, "") == 0) {
		fp = stdout;
	}
	else if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}
	if ((fp = fopen(file, "w")) == NULL) {
		fprintf(stderr, "write_output: can't open file %s\n", file);
		return(1);
	}

	for (i = 0; i < num_stars; i++) {
		if (star_arr[i].include != 0)  {
			fprintf(fp, "%6d %6.3f %6.3f\n", star_arr[i].id, 
						mag0[i + 1], sigm0[i + 1]);
		}
	}

	if (fp != stdout) {
		fclose(fp);
	}

	return(0);
}




#if 0
	/************************************************************************
	 * PROCEDURE: calc_model_mag
	 *
	 * DESCRIPTION: We have finished finding the photometric solution,
	 *              so we have the image adjustments, mean star magnitudes,
	 *              and chip offsets.  Here, we figure out the magnitude
	 *              a star _ought_ to have, based on its image, star,
	 *              chip, etc.  We go back to the x[] vector values
	 *              for all these calculations.
	 *
	 *              The value returned can be compared to the actual
	 *              measured magnitude; in a perfect world, they
	 *              would be the same.  In the real world, we expect
	 *              small residuals which (if we have done a good 
	 *              job in the model) will be randomly scattered
	 *              around zero.
	 *
	 * RETURNS:
	 *     magnitude of star from solution
	 *     NODATA         if this isn't a valid measurement
	 */

static double
calc_model_mag
	(
	int num_images,        /* I: total number of images in solution */
	int num_stars,         /* I: total number of stars in solution */
	int image_index,       /* I: index of image, 1 ... num_images */
	int star_index         /* I: index of star, 1 .. num_stars */
	)
{
	double star_mag0, mag_correction;
	double model_mag;

	/* we don't (yet) consider whether this is a valid measurement */

	star_mag0 = x[num_images + star_index];
	mag_correction = calc_model_correction(num_images, num_stars,
			              image_index, star_index);

	if (mag_correction == NODATA) {
		model_mag = NODATA;
	}
	else {
		model_mag = star_mag0 + mag_correction;
	}

	return(model_mag);
}

#endif



	/************************************************************************
	 * PROCEDURE: calc_model_correction
	 
	 * DESCRIPTION: We have finished finding the photometric solution,
	 *              so we have the image adjustments, mean star magnitudes,
	 *              and chip offsets.  Here, we figure out the total
	 *              size of all the adjustments, due to image zeropoint,
	 *              chip offset, filter mismatch, etc.
	 *
	 *              The result should be used in this sense:
	 *
	 *                  obs_mag  =   true_mag  +  corrections
	 *
	 *
	 * RETURNS:
	 *     corrections to star's mag
	 *     NODATA       if we can't calculate it
	 */

static double
calc_model_correction
	(
	int num_images,        /* I: total number of images in solution */
	int num_stars,         /* I: total number of stars in solution */
	int image_index,       /* I: index of image, 1 ... num_images */
	int star_index         /* I: index of star, 1 .. num_stars */
	)
{
	double image_em;
	double mag_correction;

	/* we don't (yet) consider whether this is a valid measurement */
	image_em = x[image_index];

	mag_correction = image_em;

	return(mag_correction);
}




