

	/*************************************************************************
	 * Take as input a file which contains the instrumental photometry from
	 * many SNAP images, re-organized so that all data for a given star
	 * is collected together.  Set up a set of linear equations which
	 * represent the photometric model of the entire system, and then 
	 * solve those equations.  Print out final reduced differential
	 * magnitudes for each star, as well as diagnostic information about
	 * each image.
	 *
	 * MWR 10/18/2003
	 *
	 * Added command-line option "primary=" to pick out primary fiducial
	 * filter index.
	 *
	 * MWR 10/31/2003
	 *
	 * Added command-line option "fcfile=" to place .fc output into a
	 * desired file instead of "solvesnap.fc".
	 *  
	 * MWR 8/6/2004
	 *
	 * Reverted to the format of the old "solvepht", which expects
	 * simply a list of stellar measurements from different exposures
	 * of a single chip.  Took out all the SNAP-specific stuff to
	 * make a simple, general tool.
	 * MWR 9/19/2004
	 *
	 * Fixed the "markoff()" procedure so that it correctly
	 * compares the ID value given by user on command-line
	 * to the ID values of stars (not their index positions).
	 * Also make sure that we set variable field to 1 for
	 * varstar= stars (in some old versions, we set to 0, making
	 * variables _not_ variable!).
	 * MWR 10/26/2004
	 */

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include "solvepht.h"

#define IMFILE     "solvepht.img"
#define SIGFILE    "solvepht.sig"
#define OUTFILE    "solvepht.out"

static char *progname = "solvepht";


static int markoff(int nbadim, int *badim, int nbadstar, int *badstar, 
                    int nvarstar, int *varstar);


char usage[] = "usage: solvepht [debug] [infile=] [mconst=] [outfile=] [imfile=] [sigfile=] \n                [badim=] [badstar=] [varstar=] [min=] [max=] \n";

int
main
	(
	int argc, 
	char *argv[]
	)
{
	int i, nn, nbadim, nbadstar, nvarstar, *badim, *badstar, *varstar;
	char infile[100], outfile[100], imfile[100], sigfile[100];
	double mconst, min, max;

	/* set some global variables to initial known values */
	image_arr = NULL;
	star_arr = NULL;
	measure_list = NULL;

	debug_flag = 0;
	nbadim = 0;
	nbadstar = 0;
	nvarstar = 0;
	min = -1.0e6;
	max = 1.0e6;
	if ((badim = (int *) calloc(argc, sizeof(int))) == NULL) {
		fprintf(stderr, "%s: can't malloc for %d ints \n", progname, argc);
		exit(1);
	}
	if ((badstar = (int *) calloc(argc, sizeof(int))) == NULL) {
		fprintf(stderr, "%s: can't malloc for %d ints \n", progname, argc);
		exit(1);
	}
	if ((varstar = (int *) calloc(argc, sizeof(int))) == NULL) {
		fprintf(stderr, "%s: can't malloc for %d ints \n", progname, argc);
		exit(1);
	}
	mconst = MAG_CONST;
	strcpy(infile, "");
	strcpy(outfile, OUTFILE);
	strcpy(imfile, IMFILE);
	strcpy(sigfile, SIGFILE);

	for (i = 1; i < argc; i++) {
		if (strcmp(argv[i], "help") == 0) {
			fprintf(stderr, "%s\n", usage);
			exit(0);
		}
		if (strcmp(argv[i], "debug") == 0) {
			debug_flag = 1;
			continue;
		}
		if (strncmp(argv[i], "infile=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%s", infile) != 1) {
				fprintf(stderr, "%s: bad value for infile ..%s..\n", 
						              progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "outfile=", 8) == 0) {
			if (sscanf(argv[i] + 8, "%s", outfile) != 1) {
				fprintf(stderr, "%s: bad value for outfile ..%s..\n", 
						              progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "imfile=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%s", imfile) != 1) {
				fprintf(stderr, "%s: bad value for imfile ..%s..\n", 
						              progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "sigfile=", 8) == 0) {
			if (sscanf(argv[i] + 8, "%s", sigfile) != 1) {
				fprintf(stderr, "%s: bad value for sigfile ..%s..\n", 
						               progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "mconst=", 7) == 0) {
			if (sscanf(argv[i] + 7, "%lf", &mconst) != 1) {
				fprintf(stderr, "%s: bad value for mconst ..%s..\n", 
						               progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "badim=", 6) == 0) {
			if (sscanf(argv[i] + 6, "%d", &nn) != 1) {
				fprintf(stderr, "%s: bad value for badim ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			badim[nbadim++] = nn;
			continue;
		}
		if (strncmp(argv[i], "badstar=", 8) == 0) {
			if (sscanf(argv[i] + 8, "%d", &nn) != 1) {
				fprintf(stderr, "%s: bad value for badstar ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			badstar[nbadstar++] = nn;
			continue;
		}
		if (strncmp(argv[i], "varstar=", 8) == 0) {
			if (sscanf(argv[i] + 8, "%d", &nn) != 1) {
				fprintf(stderr, "%s: bad value for varstar ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			varstar[nvarstar++] = nn;
			continue;
		}
		if (strncmp(argv[i], "min=", 4) == 0) {
			if (sscanf(argv[i] + 4, "%lf", &min) != 1) {
				fprintf(stderr, "%s: bad value for min ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		if (strncmp(argv[i], "max=", 4) == 0) {
			if (sscanf(argv[i] + 4, "%lf", &max) != 1) {
				fprintf(stderr, "%s: bad value for max ..%s..\n", 
						                progname, argv[i]);
				exit(1);
			}
			continue;
		}
		fprintf(stderr, "%s\n", usage);
		exit(1);
	}

	if (read_input(infile, min, max) != 0) {
		fprintf(stderr, "%s: read_input returns with error \n", progname);
		exit(1);
	}

	if (markoff(nbadim, badim, nbadstar, badstar, nvarstar, varstar) != 0) {
		fprintf(stderr, "%s: markoff returns with error \n", progname);
		exit(1);
	}

	/* 
	 * do I need to run "good_stars()" and "good_images()" again, now that
	 *   all the "bad" images and stars have been marked as such?  
	 *   Think about it.
	 *   MWR 12/3/2003.
	 */


	if (debug_flag) {
		print_images(stdout);
		print_stars(stdout, LONG);
	}

	solve_matrix(mconst);

	write_images(imfile);
	write_sigma(sigfile);
	write_stars(outfile);

	exit(0);
}



	/***********************************************************************
	 * PROCEDURE: markoff
	 *
	 * DESCRIPTION: mark all the given stars and/or images as BAD, 
	 *              which means they're not used in the solution at all.
	 *              Go through all measurements associated with these
	 *              images and stars, and set the measurements'
	 *              "include" field to zero as well.
	 *
	 *              Mark as VARIABLE the stars which the user flagged
	 *              as "varstar" on the command line. VARIABLE stars are
	 *              are not used to solve for the em's, but later
	 *              do have their magnitudes calculated. 
	 *
	 * RETURNS:
	 *     0        if all goes well
	 *     1        if an error occurs
	 */

static int 
markoff
	(
	int nbadim, 
	int *badim, 
	int nbadstar, 
	int *badstar, 
	int nvarstar, 
	int *varstar
	)
{
	int i, j;
	int max_id;
	int num_removed;
	int num_good_removed;
	struct s_image *image;
	struct s_star *star;
	struct s_measurep *mp;
	struct s_measure *measure;


	/* first, we set to 0 the "include" field of all bad images */
	for (i = 0; i < nbadim; i++) {
		if ((badim[i] < 0) || (badim[i] >= num_images)) {
			fprintf(stderr, "%s: markoff: nbadim = %d, badim[%d] = %d? \n", 
					progname, nbadim, i, badim[i]);
			return(1);
		}
		image_arr[badim[i]].include = 0;
	}


	/* 
	 * next, we set to 0 the "include" field of all the measurements
	 *   which occurred in those images.
	 */
	num_removed = 0;
	num_good_removed = 0;

	for (i = 0; i < num_images; i++) {
		image = &(image_arr[i]);
		if (image->include == 0) {
			mp = image->mp;
			while (mp != NULL) {
				shAssert((measure = mp->measure) != NULL);
				if (measure->include != 0) {
					num_good_removed++;
				}
				num_removed++;
				measure->include = 0;
				mp = mp->next;
			}
		}
	}
#ifdef DEBUG
	printf("markoff: img  %6d removed  %6d good removed \n", 
					num_removed, num_good_removed);
#endif


	/* 
	 * now, it is _possible_ that there may be some stars
	 * with ID numbers larger than the total number of stars;
	 * the user may have edited the list of images and stars
	 * provided by "multipht".  So, we make a special search
	 * here for the maximum star ID number.
	 */
	max_id = -1;
	for (i = 0; i < num_stars; i++) {
		if (star_arr[i].id > max_id) {
			max_id = star_arr[i].id;
		}
	}
#ifdef DEBUG
	printf("markoff: num_stars %d  max star ID number %d \n", 
						num_stars, max_id);
#endif


	/* 
	 * now, we go through the list of all "bad" star ID values
	 * and find those ID values in the array of stars.
	 * We'll mark those entries not to be included in calculations.
	 */
	for (i = 0; i < nbadstar; i++) {
		if ((badstar[i] < 0) || (badstar[i] > max_id)) {
			fprintf(stderr, "%s: markoff: nbadstar = %d, badstar[%d] = %d? \n", 
					progname, nbadstar, i, badstar[i]);
			return(1);
		}
		for (j = 0; j < num_stars; j++) {
			if (star_arr[j].id == badstar[i]) {
				star_arr[j].include = 0;
				break;
			}
		}
	}
	/* 
	 * and we also need to set to 0 the "include" field of all 
	 * the _measurements_ of these stars
	 */
	num_removed = 0;
	num_good_removed = 0;
	for (i = 0; i < num_stars; i++) {
		star = &(star_arr[i]);
		if (star->include == 0) {
			mp = star->mp;
			while (mp != NULL) {
				shAssert((measure = mp->measure) != NULL);
				if (measure->include != 0) {
					num_good_removed++;
				}
				num_removed++;
				measure->include = 0;
				mp = mp->next;
			}
		}
	}
#ifdef DEBUG
	printf("markoff: star %6d removed  %6d good removed \n", 
					num_removed, num_good_removed);
#endif
		

	/* 
	 * finally, we set the "variable" field to 1 for all stars
	 *   which the user flagged as "varstar" in the command line 
	 */

	for (i = 0; i < nvarstar; i++) {
		if ((varstar[i] < 0) || (varstar[i] > max_id)) {
			fprintf(stderr, "%s: markoff: nvarstar = %d, varstar[%d] = %d? \n", 
					progname, nvarstar, i, varstar[i]);
			return(1);
		}
		for (j = 0; j < num_stars; j++) {
			if (star_arr[j].id == varstar[i]) {
				star_arr[j].variable = 1;
				break;
			}
		}
	}

	return(0);
}


